/* -*- compile-command: "msdev trace.dsp /make \"Trace - Win32 Debug\"" -*-
 */

#include "engine.h"
#include "read.h"
#include "scene.h"
#include "general.h"
#include "trace.h"
#include <process.h>


extern "C" {
    int __declspec( dllexport ) Trace_Init( Tcl_Interp* interp );
    BOOL __declspec( dllexport ) WINAPI DllMain( HINSTANCE, DWORD, LPVOID );

	int GetScene( ClientData, Tcl_Interp*, int, Tcl_Obj* const [] );
	int Trace_ResizeCmd( ClientData clientData, Tcl_Interp* interp, int objc, Tcl_Obj* const objv[] );
	int Trace_RedrawCmd( ClientData clientData, Tcl_Interp* interp, int objc, Tcl_Obj* const objv[] );
	int Start_Trace_Render( ClientData clientData, Tcl_Interp* interp, int objc, Tcl_Obj* const objv[] );
	int SaveBitmap(ClientData clientData, Tcl_Interp* interp, int objc, Tcl_Obj* const objv[]);
	int StopThread(ClientData clientData, Tcl_Interp* interp, int objc, Tcl_Obj* const objv[]);
	void Trace_Render( void *dummy);
}


enum Status_e {RUNNING, STOPPED, HALT};

HINSTANCE g_hInstance = NULL;
Tcl_Interp* g_interp = NULL;
int g_height;
int g_width;
Tcl_Obj* g_pWidth = NULL;
Tcl_Obj* g_pHeight = NULL;
unsigned char* g_picture;
static Status_e threadStatus = STOPPED;
static HANDLE threadMutex;


char command_buffer[1024];
char convert_buffer[1024];

Scene *scene;

/* DllMain
 * 
 * every DLL must provide DllMain.  ours just stores the instance handle,
 * so we can use it to create a window later.
 */

BOOL WINAPI DllMain( HINSTANCE hInst, DWORD dwReason, LPVOID reserved )
{
    g_hInstance = hInst;
    return 1;
}


void refreshScreen(void)
{
	glClearColor(0.0, 0.0, 0.0, 0.0);

	glClear(GL_COLOR_BUFFER_BIT);
	
	if( WaitForSingleObject(threadMutex,1000) == WAIT_TIMEOUT ) return;
		

	if (g_picture != NULL)
	{
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

		glDrawPixels(g_width, g_height, GL_RGB, GL_UNSIGNED_BYTE, g_picture);

		glFlush();
	}
	ReleaseMutex(threadMutex);
}

int StopThread(ClientData clientData, Tcl_Interp* interp, int objc, Tcl_Obj* const objv[])
{
    if( WaitForSingleObject(threadMutex, 1000) == WAIT_TIMEOUT) return TCL_OK;

	if(threadStatus == RUNNING)
	{
		threadStatus = HALT;
		while(threadStatus != STOPPED) Sleep(10);
	}
	ReleaseMutex(threadMutex);
	return TCL_OK;
}



int Start_Trace_Render( ClientData clientData, Tcl_Interp* interp, int objc, Tcl_Obj* const objv[])
{
	
	if(scene != NULL)
	{	
		
		if( WaitForSingleObject(threadMutex,1000) == WAIT_TIMEOUT ) 
			return TCL_OK;

	
		if( Tcl_GetIntFromObj( interp, Tcl_ObjGetVar2( interp, g_pWidth, NULL, TCL_GLOBAL_ONLY),
								&g_width) != TCL_OK)
		{
			ReleaseMutex(threadMutex);
			return TCL_OK;
		}
		if( Tcl_GetIntFromObj( interp, Tcl_ObjGetVar2( interp, g_pHeight, NULL, TCL_GLOBAL_ONLY),
								&g_height) != TCL_OK)
		{
			ReleaseMutex(threadMutex);
			return TCL_OK;
		}
		threadStatus = RUNNING;
		delete [] g_picture;
		g_picture = new GLubyte [g_width * g_height * 3];
		memset(g_picture, 0, g_width*g_height*3);
		_beginthread( &Trace_Render , 0, NULL);
		ReleaseMutex(threadMutex);
	}
	return TCL_OK;
}


void Trace_Render( void *dummy )
{

	char buff[] = {".gl redraw"};
	RayTracer ray_tracer;

	for (unsigned int width_index = 0; width_index < g_width; ++width_index)
	{
		for (unsigned int height_index = 0; height_index < g_height; ++height_index)
		{	
			if(threadStatus == HALT) 
			{
				threadStatus = STOPPED;	//We allow this because we know the mutex is held in StopThread
				_endthread();			//and it is waiting for us.  
			}

            double offset_x = 0.5*((static_cast<double>(g_height) - static_cast<double>(g_width))/static_cast<double>(g_height));
			double normalized_y = static_cast<double>(height_index) / static_cast<double>(g_height);
			double normalized_x = offset_x + (static_cast<double>(width_index) / static_cast<double>(g_height));

			vec3 color = ray_tracer.trace(scene, normalized_x, normalized_y);

			if( WaitForSingleObject(threadMutex,100) == WAIT_TIMEOUT )
			{

				height_index = height_index - 1;
				width_index = width_index - 1;
				continue;
			}

			g_picture[(g_width * height_index + width_index) * 3 + 0] = color[0] * 255;//red;
			g_picture[(g_width * height_index + width_index) * 3 + 1] = color[1] * 255;//green;
			g_picture[(g_width * height_index + width_index) * 3 + 2] = color[2] * 255;//blue;
			
            ReleaseMutex(threadMutex);
		}
	}

	while(WaitForSingleObject(threadMutex,1000) == WAIT_TIMEOUT) continue;
	threadStatus = STOPPED;
	ReleaseMutex(threadMutex);
		

	return;
}

int Trace_ResizeCmd( ClientData clientData, Tcl_Interp* interp, int objc, Tcl_Obj* const objv[] )
{

	refreshScreen();


  return TCL_OK;
}


int Trace_RedrawCmd( ClientData clientData, Tcl_Interp* interp, int objc, Tcl_Obj* const objv[] )
{
	refreshScreen();

	return TCL_OK;
}

/* Trace_Init
 *
 * Tcl calls this function when the DLL is first loaded, to initialize
 * the package(s) contained in the DLL.
 */

int Trace_Init( Tcl_Interp* interp )
{
	Tcl_CreateObjCommand( interp, "trace::getscene", GetScene, NULL, NULL);
	Tcl_CreateObjCommand( interp, "trace::resize", Trace_ResizeCmd, NULL, NULL ); 
	Tcl_CreateObjCommand( interp, "trace::redraw", Trace_RedrawCmd, NULL, NULL ); 
	Tcl_CreateObjCommand( interp, "trace::render", Start_Trace_Render, NULL, NULL ); 
	Tcl_CreateObjCommand( interp, "trace::save", SaveBitmap, NULL, NULL ); 
	Tcl_CreateObjCommand( interp, "trace::stoprender", StopThread, NULL, NULL ); 
    
    /* initialize the UI control package. */
    if ( Controls_Init( interp ) != TCL_OK )
	return TCL_ERROR;

    /* keep a pointer to the interpreter. */
    g_interp = interp;
	threadMutex = CreateMutex( NULL, FALSE, "mutex - guards picture" );
	g_pWidth = Tcl_NewStringObj("g_width", -1);
	g_pHeight = Tcl_NewStringObj("g_height", -1);

    return TCL_OK;
}

/* Debug
 *
 * Debug works just like printf(), except that it prints output in the
 * MSVC++ "output" window.  don't try to print more than 1K bytes at
 * once.
 */

void Debug( char* fmt, ... )
{
    char buffer[1024];
    va_list list;

    va_start( list, fmt );
    vsprintf( buffer, fmt, list );
    va_end( list );

    OutputDebugString( buffer );
}



    
int SaveBitmap(ClientData clientData, Tcl_Interp* interp,
			  int objc, Tcl_Obj* const objv[])
{
		FILE* file;
		OPENFILENAME open;
		char fname[100];
		
		BITMAPFILEHEADER bmfh;
		BITMAPINFOHEADER bmih;

		fname[0] = '\0';

		open.lStructSize = sizeof(OPENFILENAME);
		open.hwndOwner = NULL;
		open.lpstrFilter = "Bitmap (*.bmp)\0*.bmp\0\0";
		open.nFilterIndex = 0;
		open.lpstrCustomFilter = NULL;
		open.nFilterIndex = NULL;
		open.lpstrFile = fname;
		open.nMaxFile = 2048000;
		open.lpstrFileTitle = NULL;
		open.lpstrInitialDir = NULL;
		open.lpstrTitle = "Bmp loader";
		open.Flags = NULL;
		open.nFileOffset = NULL;
		open.nFileExtension = NULL;
		open.lpstrDefExt = "bmp";
		open.lCustData = NULL;
		open.lpfnHook = NULL;
		open.lpTemplateName = NULL;


		GetSaveFileName(&open);
	

		int bytes, pad;
		bytes = g_width * 3;
		pad = (bytes%4) ? 4-(bytes%4) : 0;
		bytes += pad;
		bytes *= g_height;

		bmfh.bfType = 0x4d42;    // "BM"
		bmfh.bfSize = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + bytes;
		bmfh.bfReserved1 = 0;
		bmfh.bfReserved2 = 0;
		bmfh.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);

		bmih.biSize = sizeof(BITMAPINFOHEADER);
		bmih.biWidth = g_width;
		bmih.biHeight = g_height;
		bmih.biPlanes = 1;
		bmih.biBitCount = 24;
		bmih.biCompression = BI_RGB;
		bmih.biSizeImage = 0;
		bmih.biXPelsPerMeter = (int)(100 / 2.54 * 72);
		bmih.biYPelsPerMeter = (int)(100 / 2.54 * 72);
		bmih.biClrUsed = 0;
		bmih.biClrImportant = 0;

		file = fopen(open.lpstrFile, "wb");
		
		fwrite( &bmfh, sizeof(BITMAPFILEHEADER), 1, file );
		fwrite( &bmih, sizeof(BITMAPINFOHEADER), 1, file );

		bytes /= g_height;
		unsigned char* scanline = new unsigned char [bytes];
		for ( int j = 0; j < g_height; ++j )
		{
			memcpy( scanline, g_picture + j*3*g_width, bytes );
			for ( int i = 0; i < g_width; ++i )
			{
				unsigned char temp = scanline[i*3];
				scanline[i*3] = scanline[i*3+2];
				scanline[i*3+2] = temp;
			}
			fwrite( scanline, bytes, 1, file );
		}
		delete [] scanline;
		fclose(file);
	return TCL_OK;
}
