/* -*- compile-command: "msdev trace.dsp /make \"Trace - Win32 Debug\"" -*-
 */

#ifndef _ENGINE_H
#define _ENGINE_H

#include <windows.h>
#include <tk.h>
#include <assert.h>

#include "gl\gl.h"
#include "gl\glu.h"

#define MAX_CONTROL_NAME  80
#define MAX_MENU_ITEMS    20


#define PICTURE_WIDTH 1
#define PICTURE_HEIGHT 2
#define REC_DEPTH 3
#define REC_THRESH 4

extern unsigned char* g_picture;
extern int g_width;
extern int g_height;

typedef struct
{
    int type;
    Tcl_Obj* widget;
    int curveid;
    void* data;
} control;

typedef struct
{
    double min, max, res, start;
    char name[MAX_CONTROL_NAME+1];
} scale_data;


typedef struct
{
    int state;
    char name[MAX_CONTROL_NAME+1];
} checkbox_data;

typedef struct
{
    int start;
    int count;
    char name[MAX_CONTROL_NAME+1];
    char item[MAX_MENU_ITEMS][MAX_CONTROL_NAME+1];
} menu_data;



/* grrr.  why does Microsoft not put the value of pi in any of its
   standard include files? */
#define M_PI 3.14159265358979323846

/* we'll use static shared buffers for constructing Tcl commands
   (which are evaluated and immediately thrown away). */
#define BUFFER_SIZE  1024
extern char command_buffer[BUFFER_SIZE];
extern char convert_buffer[BUFFER_SIZE];


extern Tcl_Interp* g_interp;


/* CONVERT (used in controls.c) takes the string variable "s" and
   escapes spaces and funny characters in it according to Tcl rules,
   placing the result in convert_buffer. */
#define CONVERT(s) {int flags=0; \
                   assert( Tcl_ScanElement( s, &flags ) < BUFFER_SIZE );\
                   Tcl_ConvertElement( s, convert_buffer, flags );}
		   
/* g_quality controls polygonalization of the sphere and cylinder
   primitives.
   higher g_quality --> more polygons --> prettier but slower. */
extern int g_quality;


LRESULT CALLBACK WndProc( HWND hwnd, UINT iMsg,
			  WPARAM wParam, LPARAM lParam );
HGLRC SetupOpenGL( HWND hwnd );


/* Debug is like printf, but prints to the MSVC debugger "output"
   window. */
void __declspec( dllexport ) Debug( char* fmt, ... );
double __declspec( dllexport ) get_time( void );




/** user interface widgets **/

/* internal stuff, can't call these. */
int Controls_Init( Tcl_Interp* interp );
control* new_control( int, int );

#define CONTROL_SCALE        1
#define CONTROL_CHECKBOX     2
#define CONTROL_RADIOBUTTON  3
#define CONTROL_MENU         4

/* for all controls, interp is the Tcl_Interp parameter passed to
   Trace_SetupCmd, id is an integer unique to that control, and name
   is a string used for the label. */


/* slider from "min" to "max" in steps of "res", initial value "start". */
int __declspec( dllexport )
    scale( Tcl_Interp* interp, int id, char* name, double min, double max,
	   double res, double start );
void scale_widget( control* );

/* checkbox, initial state "state". */
int __declspec( dllexport )
    checkbox( Tcl_Interp* interp, int id, char* name, int state );
void checkbox_widget( control* );

/* set of radiobuttons with labels "..." (NULL-terminated).  "start" is
   the number of the one initially checked. */
int __declspec( dllexport )
    radiobutton( Tcl_Interp* interp, int id, char* name, int start, ... );
void radiobutton_widget( control* );

/* menu with radiobutton-type options "..." (NULL-terminated).
   "start" is the one initially checked. */
int __declspec( dllexport )
    menu( Tcl_Interp* interp, int id, char* name, int start, ... );
void menu_widget( control* );
    
/* read the value of a control as a double, int, or boolean. */    
double __declspec( dllexport ) get_control_d( int id );
int __declspec( dllexport ) get_control_i( int id );
int __declspec( dllexport ) get_control_b( int id );




#endif 


    
